import argparse
import os

import dataloader
import jax
import matplotlib.pyplot as plt
import numpy as np
import optax
import wandb
from absl import logging
from flax import nnx
from jax import random as jr
from jax.experimental import mesh_utils

import blaxbird
from blaxbird import get_default_checkpointer, train_fn
from blaxbird.experimental import rfm


def get_optimizer(model, lr=1e-4):
  tx = optax.adamw(lr)
  tx = nnx.Optimizer(model, tx=tx)
  return tx


def get_sharding():
  num_devices = jax.local_device_count()
  mesh = jax.sharding.Mesh(
    mesh_utils.create_device_mesh((num_devices,)), ("data",)
  )
  model_sharding = jax.NamedSharding(mesh, jax.sharding.PartitionSpec())
  data_sharding = jax.NamedSharding(mesh, jax.sharding.PartitionSpec("data"))
  return model_sharding, data_sharding


def visualize_hook(sample_fn, val_iter, hook_every_n_steps, log_to_wandb):
  n_row, n_col, img_size = 12, 32, (32, 32, 3)

  def convert_batch_to_image_grid(image_batch):
    reshaped = (
      image_batch.reshape(n_row, n_col, *img_size)
      .transpose([0, 2, 1, 3, 4])
      .reshape(n_row * img_size[0], n_col * img_size[1], img_size[2])
    )
    return (reshaped + 1.0) / 2.0

  def plot(images):
    fig = plt.figure(figsize=(16, 6))
    ax = fig.add_subplot(1, 1, 1)
    ax.imshow(
      images,
      interpolation="nearest",
      cmap="gray",
    )
    plt.axis("off")
    plt.tight_layout()
    return fig

  def fn(step, *, model, **kwargs):
    if step % hook_every_n_steps != 0:
      return
    all_samples = []
    for i, batch in enumerate(val_iter):
      samples = sample_fn(
        model, jr.fold_in(jr.key(step), i), sample_shape=batch["inputs"].shape
      )
      all_samples.append(samples)
      if len(all_samples) * all_samples[0].shape[0] >= n_row * n_col:
        break
    all_samples = np.concatenate(all_samples, axis=0)[: (n_row * n_col)]
    all_samples = convert_batch_to_image_grid(all_samples)
    fig = plot(all_samples)
    if jax.process_index() == 0 and log_to_wandb:
      wandb.log({"images": wandb.Image(fig)}, step=step)

  return fn


def get_hooks(sample_fn, val_itr, hook_every_n_steps, log_to_wandb):
  return [visualize_hook(sample_fn, val_itr, hook_every_n_steps, log_to_wandb)]


def get_train_and_val_itrs(rng_key, outfolder):
  return dataloader.data_loaders(
    rng_key,
    outfolder,
    split=["train[:90%]", "train[90%:]"],
    shuffle=[True, False],
  )


def run(n_steps, eval_every_n_steps, n_eval_batches, dit_type, log_to_wandb):
  logging.set_verbosity(logging.INFO)
  outfolder = os.path.join(os.path.dirname(__file__), "workdir")

  train_itr, val_itr = get_train_and_val_itrs(
    jr.key(0), os.path.join(outfolder, "data")
  )

  model = getattr(blaxbird.experimental, dit_type)(
    image_size=(32, 32, 3), rngs=nnx.rnglib.Rngs(jr.key(1))
  )
  train_step, val_step, sample_fn = rfm()
  optimizer = get_optimizer(model)

  save_fn, _, restore_last_fn = get_default_checkpointer(
    os.path.join(outfolder, "checkpoints"),
    save_every_n_steps=eval_every_n_steps,
  )
  hooks = get_hooks(sample_fn, val_itr, eval_every_n_steps, log_to_wandb) + [
    save_fn
  ]

  model_sharding, data_sharding = get_sharding()
  model, optimizer = restore_last_fn(model, optimizer)

  train = train_fn(
    fns=(train_step, val_step),
    n_steps=n_steps,
    eval_every_n_steps=eval_every_n_steps,
    n_eval_batches=n_eval_batches,
    shardings=(model_sharding, data_sharding),
    hooks=hooks,
    log_to_wandb=False,
  )
  train(jr.key(2), model, optimizer, train_itr, val_itr)


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--n-steps", type=int, default=1_000)
  parser.add_argument("--eval-every-n-steps", type=int, default=50)
  parser.add_argument("--n-eval-batches", type=int, default=10)
  parser.add_argument(
    "--dit", type=str, choices=["SmallDiT", "BaseDiT"], default="SmallDiT"
  )
  parser.add_argument("--log-to-wandb", action="store_true")
  args = parser.parse_args()
  run(
    args.n_steps,
    args.eval_every_n_steps,
    args.n_eval_batches,
    args.dit,
    args.log_to_wandb,
  )



from jax import numpy as jnp


def timestep_embedding(timesteps, embedding_dim: int, dtype=jnp.float32):
  half = embedding_dim // 2
  freqs = jnp.exp(-jnp.log(10_000) * jnp.arange(0, half) / half)
  emb = timesteps.astype(dtype)[:, None] * freqs[None, ...]
  emb = jnp.concatenate([jnp.sin(emb), jnp.cos(emb)], axis=1)
  return emb



import chex
import jax
import numpy as np
from flax import nnx
from jax import numpy as jnp
from jax import random as jr

from blaxbird._src.experimental.edm import EDMConfig
from blaxbird._src.experimental.rfm import (
  RFMConfig,
)


def euler_sample_fn(config: RFMConfig):
  """Construct an Euler sampler for flow matching.

  Args:
    config: a FlowMatchingConfig object

  Returns:
    returns a callable that can be used to sample from a flow matching model
  """

  def sample_fn(
    model: nnx.Module,
    rng_key: jax.Array,
    sample_shape: tuple = (),
    *,
    context: jax.Array = None,
  ) -> jax.Array:
    """Sample from a flow matching model.

    Args:
      model: a nnx.Module that is used as the learned vector field in flow
       matching
      rng_key: a jax.random.key object
      sample_shape: the shape of the data to be generated, where the first axis
        is the batch dimension and the other axes are the feature dimensions
      context: a conditioning variable (if used)

    Returns:
      returns a sample from the model
    """
    if context is not None:
      chex.assert_equal(sample_shape[0], len(context))
    dt = 1.0 / config.n_sampling_steps
    samples = jr.normal(rng_key, sample_shape)
    time_steps = config.parameterization.sampling_sigmas(
      config.n_sampling_steps
    )
    for times in time_steps:
      times = jnp.repeat(times, samples.shape[0])  # noqa: PLW2901
      vt = model(inputs=samples, times=times, context=context)
      samples = samples + vt * dt
    return samples

  return sample_fn


def heun_sampler_fn(config: EDMConfig):
  """Construct a Heun sampler for denoising score matching.

  Args:
    config: a EDMConfig object

  Returns:
    returns a callable that can be used to sample from a score matching model
  """
  params = config.parameterization

  # ruff: noqa: ANN001, ANN202, ANN003
  def _denoise(model, rng_key, inputs, sigma, context):
    new_shape = (-1,) + tuple(np.ones(inputs.ndim - 1, dtype=np.int32).tolist())
    inputs_t = inputs * params.in_scaling(sigma).reshape(new_shape)
    noise_cond = params.noise_conditioning(sigma)
    outputs = model(
      inputs=inputs_t,
      context=context,
      times=noise_cond,
    )
    skip = inputs * params.skip_scaling(sigma).reshape(new_shape)
    outputs = outputs * params.out_scaling(sigma).reshape(new_shape)
    outputs = skip + outputs
    return outputs

  def sample_fn(
    model: nnx.Module,
    rng_key: jax.Array,
    sample_shape: tuple = (),
    *,
    context: jax.Array = None,
  ) -> jax.Array:
    """Sample from a score matching model.

    Args:
      model: a nnx.Module that is used as the learned score model in score
        matching
      rng_key: a jax.random.key object
      sample_shape: the shape of the data to be generated, where the first axis
        is the batch dimension and the other axes are the feature dimensions
      context: a conditioning variable (if used)

    Returns:
      returns a sample from the model
    """
    if context is not None:
      chex.assert_equal(sample_shape[0], len(context))
    n = context.shape[0]
    noise_key, rng_key = jr.split(rng_key)
    sigmas = params.sampling_sigmas(config.n_sampling_steps)
    samples = jr.normal(rng_key, sample_shape) * sigmas[0]

    for i, (sigma, sigma_next) in enumerate(zip(sigmas[:-1], sigmas[1:])):
      pred_key1, pred_key2, rng_key = jr.split(rng_key, 3)
      sample_curr = samples
      pred_curr = _denoise(
        model,
        pred_key1,
        inputs=sample_curr,
        sigma=jnp.repeat(sigma, n),
        context=context,
      )
      d_cur = (sample_curr - pred_curr) / sigma
      samples = sample_curr + d_cur * (sigma_next - sigma)
      # second order correction
      if i < config.n_sampling_steps - 1:
        pred_next = _denoise(
          model,
          pred_key2,
          inputs=samples,
          sigma=jnp.repeat(sigma_next, n),
          context=context,
        )
        d_prime = (samples - pred_next) / sigma_next
        samples = sample_curr + (sigma_next - sigma) * (
          0.5 * d_cur + 0.5 * d_prime
        )
    return samples

  return sample_fn



import os
from collections.abc import Callable

import orbax.checkpoint as ocp
from absl import logging
from flax import nnx


def get_default_checkpointer(
  outfolder: str,
  *,
  save_every_n_steps: int,
  max_to_keep: int = 5,
) -> tuple[Callable, Callable, Callable]:
  """Construct functions for checkpointing functionality.

  Args:
    outfolder: a path specifying where checkpoints are stored
    save_every_n_steps: how often to store checkpoints
    max_to_keep: number of checkpoints to store before they get deleted

  Returns:
    returns function to saev and restore checkpoints
  """
  checkpointer = ocp.PyTreeCheckpointer()
  options = ocp.CheckpointManagerOptions(
    max_to_keep=max_to_keep,
    create=True,
    best_mode="min",
    best_fn=lambda x: x["val/loss"],
  )
  checkpoint_manager = ocp.CheckpointManager(
    os.path.join(outfolder, "best"),
    options=options,
    item_names=("state", "opt_state"),
  )

  def save_fn(
    step: int, *, model: nnx.Module, optimizer: nnx.Optimizer, metrics: dict
  ) -> None:
    """Save a checkpoint.

    Args:
       step: number of training/gradient step
       model: a nnx.Model object
       optimizer: a nnx.Optimizer object
       metrics: a dictionary of metrics. One of which must be the criterion
        specified above.
    """
    if step % save_every_n_steps != 0:
      return

    _, state = nnx.split(model)
    _, opt_state = nnx.split(optimizer.opt_state)
    try:
      logging.info("saving best checkpoint")
      checkpoint_manager.save(
        step=step,
        args=ocp.args.Composite(
          state=ocp.args.StandardSave(state),
          opt_state=ocp.args.StandardSave(opt_state),
        ),
        metrics=metrics,
        force=True,
      )
      checkpoint_manager.wait_until_finished()
    except Exception as e:
      logging.error(f"could not save better checkpoint because of: {e}")
      logging.error("resuming nonetheless")
    try:
      logging.info("saving last checkpoint")
      checkpointer.save(
        os.path.join(outfolder, "last"), (state, opt_state), force=True
      )
    except Exception as e:
      logging.error(f"could not last checkpoint because of: {e}")
      logging.error("resuming nonetheless")

  def restore_best_fn(
    model: nnx.Module, optimizer: nnx.Optimizer
  ) -> tuple[nnx.Module, nnx.Optimizer]:
    """Restore the best checkpoint.

    Args:
     model: a nnx.Model object
     optimizer: a nnx.Optimizer object
    """
    graph_def, state = nnx.split(model)
    opt_def, opt_state = nnx.split(optimizer.opt_state)
    try:
      logging.info("trying to restore best checkpoint")
      restored = checkpoint_manager.restore(
        checkpoint_manager.best_step(),
        args=ocp.args.Composite(
          state=ocp.args.StandardRestore(nnx.eval_shape(lambda: state)),
          opt_state=ocp.args.StandardRestore(nnx.eval_shape(lambda: opt_state)),
        ),
      )
      model = nnx.merge(graph_def, restored["state"])
      optimizer.opt_state = nnx.merge(opt_def, restored["opt_state"])
    except FileNotFoundError:
      logging.warning("could not find checkpoint. resuming with blank state")
    return model, optimizer

  def restore_last_fn(
    model: nnx.Module, optimizer: nnx.Optimizer
  ) -> tuple[nnx.Module, nnx.Optimizer]:
    """Restore the latest training checkpoint.

    Args:
     model: a nnx.Model object
     optimizer: a nnx.Optimizer object
    """
    graphdef, state = nnx.split(model)
    optdef, opt_state = nnx.split(optimizer.opt_state)
    state = nnx.eval_shape(lambda: state)
    opt_state = nnx.eval_shape(lambda: opt_state)
    try:
      logging.info("trying to restore last checkpoint")
      restored = checkpointer.restore(
        os.path.join(outfolder, "last"), (state, opt_state)
      )
      model = nnx.merge(graphdef, restored[0])
      optimizer.opt_state = nnx.merge(optdef, restored[1])
    except FileNotFoundError:
      logging.warning("could not find checkpoint. resuming with blank state")
    return model, optimizer

  return save_fn, restore_best_fn, restore_last_fn



"""Experimental models that might be moved to the main code base."""

from blaxbird._src.experimental.edm import EDMConfig, edm
from blaxbird._src.experimental.nn.dit import (
  BaseDiT,
  DiT,
  DiTBlock,
  LargeDiT,
  SmallDiT,
  XtraLargeDiT,
)
from blaxbird._src.experimental.nn.mlp import MLP
from blaxbird._src.experimental.rfm import (
  RFMConfig,
  rfm,
)

__all__ = [
  "edm",
  "EDMConfig",
  "rfm",
  "RFMConfig",
  #
  "DiT",
  "DiTBlock",
  "SmallDiT",
  "BaseDiT",
  "LargeDiT",
  "XtraLargeDiT",
  #
  "MLP",
]



import jax
from einops import rearrange
from flax import nnx
from jax import numpy as jnp

from blaxbird._src.experimental.nn.embedding import timestep_embedding
from blaxbird._src.experimental.nn.mlp import MLP


def _modulate(inputs, shift, scale):  # noqa: ANN001, ANN202
  return inputs * (1.0 + scale[:, None]) + shift[:, None]


def get_sinusoidal_embedding_1d(length, embedding_dim):  # noqa: ANN001, ANN202
  emb = timestep_embedding(length.reshape(-1), embedding_dim)
  return emb


def sinusoidal_init(shape, dtype):  # noqa: ANN001, ANN202
  def get_sinusoidal_embedding_2d(grid, embedding_dim):  # noqa: ANN001, ANN202
    emb_h = get_sinusoidal_embedding_1d(grid[0], embedding_dim // 2)
    emb_w = get_sinusoidal_embedding_1d(grid[1], embedding_dim // 2)
    emb = jnp.concatenate([emb_h, emb_w], axis=1)
    return emb

  _, n_h_patches, n_w_patches, embedding_dim = shape
  grid_h = jnp.arange(n_h_patches, dtype=jnp.float32)
  grid_w = jnp.arange(n_w_patches, dtype=jnp.float32)
  grid = jnp.meshgrid(grid_w, grid_h)

  grid = jnp.stack(grid, axis=0)
  grid = grid.reshape([2, 1, n_w_patches, n_h_patches])
  pos_embed = get_sinusoidal_embedding_2d(grid, embedding_dim)

  return jnp.expand_dims(pos_embed, 0)  # (1, H*W, D)


class OutProjection(nnx.Module):
  def __init__(  # noqa: PLR0913
    self, hidden_size, n_embedding_features, patch_size, out_channels, *, rngs
  ):
    super().__init__()
    self.ada = nnx.Sequential(
      nnx.silu, nnx.Linear(n_embedding_features, 2 * hidden_size, rngs=rngs)
    )
    self.norm = nnx.LayerNorm(hidden_size, rngs=rngs)
    self.out = nnx.Linear(
      hidden_size, patch_size * patch_size * out_channels, rngs=rngs
    )

  def __call__(self, inputs, context):
    shift, scale = jnp.split(self.ada(context), 2, -1)
    outs = self.out(_modulate(self.norm(inputs), shift, scale))
    return outs


class DiTBlock(nnx.Module):
  def __init__(  # noqa: PLR0913
    self,
    hidden_size: int,
    n_embedding_features: int,
    *,
    n_heads: int,
    dropout_rate: float = 0.1,
    rngs: nnx.rnglib.Rngs,
  ):
    """Diffusion-Transformer block.

    Args:
      hidden_size: number of features of the hidden layers
      n_embedding_features: number o features of time embedding
      n_heads: number of transformer heads
      dropout_rate: float
      rngs: random keys
    """
    super().__init__()
    self.ada = nnx.Sequential(
      nnx.silu, nnx.Linear(n_embedding_features, hidden_size * 6, rngs=rngs)
    )

    self.layer_norm1 = nnx.LayerNorm(
      hidden_size, use_scale=False, use_bias=False, rngs=rngs
    )
    self.self_attn = nnx.MultiHeadAttention(
      num_heads=n_heads, in_features=hidden_size, rngs=rngs, decode=False
    )
    self.layer_norm2 = nnx.LayerNorm(
      hidden_size, use_scale=False, use_bias=False, rngs=rngs
    )
    self.mlp = MLP(
      hidden_size,
      (hidden_size * 4, hidden_size),
      dropout_rate=dropout_rate,
      rngs=rngs,
    )

  def __call__(self, inputs: jax.Array, context: jax.Array) -> jax.Array:
    """Transform inputs through the DiT block.

    Args:
      inputs: input array
      context: values to condition on

    Returns:
      returns a jax.Array
    """
    hidden = inputs
    adaln_norm = self.ada(context)
    attn, gate = jnp.split(adaln_norm, 2, axis=-1)

    pre_shift, pre_scale, post_scale = jnp.split(attn, 3, -1)
    intermediate = _modulate(self.layer_norm1(hidden), pre_shift, pre_scale)
    intermediate = self.self_attn(intermediate)
    hidden = hidden + post_scale[:, None] * intermediate

    pre_shift, pre_scale, post_scale = jnp.split(gate, 3, -1)
    intermediate = _modulate(self.layer_norm2(hidden), pre_shift, pre_scale)
    intermediate = self.mlp(intermediate)
    outputs = hidden + post_scale[:, None] * intermediate

    return outputs


class DiT(nnx.Module):
  def __init__(  # noqa: PLR0913
    self,
    image_size: tuple[int, int, int],
    n_hidden_channels: int,
    patch_size: int,
    n_layers: int,
    n_heads: int,
    n_embedding_features=256,
    dropout_rate=0.0,
    *,
    rngs: nnx.rnglib.Rngs,
  ):
    """Diffusion-Transformer.

    Args:
      image_size: size of the image, e.g., (32, 32, 3)
      n_hidden_channels: number if hidden channels
      patch_size: size of each path
      n_layers: integer
      n_heads: integer
      n_embedding_features: integer
      dropout_rate: float
      rngs: random keys
    """
    self.image_size = image_size
    self.n_in_channels = image_size[-1]
    self.n_embedding_features = n_embedding_features
    self.patch_size = patch_size
    self.time_embedding = nnx.Sequential(
      nnx.Linear(n_embedding_features, n_embedding_features, rngs=rngs),
      nnx.swish,
      nnx.Linear(n_embedding_features, n_embedding_features, rngs=rngs),
      nnx.swish,
    )
    self.patchify = nnx.Conv(
      self.n_in_channels,
      n_hidden_channels,
      (patch_size, patch_size),
      (patch_size, patch_size),
      padding="VALID",
      kernel_init=nnx.initializers.xavier_uniform(),
      rngs=rngs,
    )
    self.patch_embedding = nnx.Param(
      sinusoidal_init(
        (
          1,
          image_size[0] // patch_size,
          image_size[1] // patch_size,
          n_hidden_channels,
        ),
        None,
      ),
    )
    self.dit_blocks = tuple(
      [
        DiTBlock(
          n_hidden_channels,
          n_embedding_features,
          n_heads=n_heads,
          dropout_rate=dropout_rate,
          rngs=rngs,
        )
        for _ in range(n_layers)
      ]
    )
    self.out_projection = OutProjection(
      n_hidden_channels,
      n_embedding_features,
      patch_size,
      self.n_in_channels,
      rngs=rngs,
    )

  def _patchify(self, inputs):
    n_h_patches = self.image_size[0] // self.patch_size
    n_w_patches = self.image_size[1] // self.patch_size
    hidden = self.patchify(inputs)
    outputs = rearrange(
      hidden, "b h w c -> b (h w) c", h=n_h_patches, w=n_w_patches
    )
    return outputs

  def _unpatchify(self, inputs):
    H = self.image_size[0] // self.patch_size
    W = self.image_size[1] // self.patch_size
    P = Q = self.patch_size
    hidden = jnp.reshape(inputs, (-1, H, W, P, Q, self.n_in_channels))
    outputs = rearrange(
      hidden, "b h w p q c -> b (h p) (w q) c", h=H, w=W, p=P, q=Q
    )
    return outputs

  def _embed(self, inputs):
    return inputs + jax.lax.stop_gradient(self.patch_embedding.value)

  def __call__(
    self, inputs: jax.Array, times: jax.Array, context: jax.Array = None
  ):
    """Transform inputs through the DiT.

    Args:
      inputs: input in image form
      times: one-dimensional array
      context: conditioning variable in image form

    Returns:
      returns a jax
    """
    hidden = self._patchify(inputs)
    hidden = self._embed(hidden)
    times = self.time_embedding(
      timestep_embedding(times, self.n_embedding_features)
    )

    for block in self.dit_blocks:
      hidden = block(hidden, context=times)

    hidden = self.out_projection(hidden, times)
    outputs = self._unpatchify(hidden)
    return outputs


def SmallDiT(image_size, patch_size=2, **kwargs):
  return DiT(
    image_size,
    n_hidden_channels=384,
    patch_size=patch_size,
    n_layers=12,
    n_heads=6,
    **kwargs,
  )


def BaseDiT(image_size, patch_size=2, **kwargs):
  return DiT(
    image_size,
    n_hidden_channels=768,
    patch_size=patch_size,
    n_layers=12,
    n_heads=12,
    **kwargs,
  )


def LargeDiT(image_size, patch_size=2, **kwargs):
  return DiT(
    image_size,
    n_hidden_channels=1024,
    patch_size=patch_size,
    n_layers=24,
    n_heads=16,
    **kwargs,
  )


def XtraLargeDiT(image_size, patch_size=2, **kwargs):
  return DiT(
    image_size,
    n_hidden_channels=1152,
    patch_size=patch_size,
    n_layers=28,
    n_heads=16,
    **kwargs,
  )



from collections.abc import Callable

import jax
from flax import nnx


class MLP(nnx.Module):
  # ruff: noqa: PLR0913, ANN204, ANN101
  def __init__(
    self,
    in_features: int,
    output_features: tuple[int, ...],
    *,
    kernel_init: nnx.initializers.Initializer = nnx.initializers.lecun_normal(),
    bias_init: nnx.initializers.Initializer = nnx.initializers.zeros_init(),
    use_bias: bool = True,
    dropout_rate: float | None = None,
    activation: Callable[[jax.Array], jax.Array] = jax.nn.silu,
    activate_last: bool = False,
    rngs: nnx.rnglib.Rngs,
  ):
    features = [in_features] + list(output_features)
    layers = []
    for index, (din, dout) in enumerate(zip(features[:-1], features[1:])):
      layers.append(
        nnx.Linear(
          in_features=din,
          out_features=dout,
          kernel_init=kernel_init,
          bias_init=bias_init,
          use_bias=use_bias,
          rngs=rngs,
        )
      )
    self.layers = tuple(layers)
    self.dropout_rate = dropout_rate
    self.activate_last = activate_last
    self.activation = activation
    if dropout_rate is not None:
      self.dropout_layer = nnx.Dropout(dropout_rate, rngs=rngs)

  def __call__(self, inputs: jax.Array):
    """Project inputs through the MLP.

    Args:
      inputs: jax.Array

    Returns:
      jax.Array
    """
    num_layers = len(self.layers)

    out = inputs
    for i, layer in enumerate(self.layers):
      out = layer(out)
      if i < num_layers - 1 or self.activate_last:
        if self.dropout_rate is not None:
          out = self.dropout_layer(out)
        out = self.activation(out)
    return out



import dataclasses

import numpy as np
from flax import nnx
from jax import numpy as jnp
from jax import random as jr

from blaxbird._src.experimental import samplers


@dataclasses.dataclass
class EDMParameterization:
  n_sampling_steps: int = 25
  sigma_min: float = 0.002
  sigma_max: float = 80.0
  rho: float = 7.0
  sigma_data: float = 0.5
  P_mean: float = -1.2
  P_std: float = 1.2
  S_churn: float = 40
  S_min: float = 0.05
  S_max: float = 50
  S_noise: float = 1.003

  def sigma(self, eps):
    return jnp.exp(eps * self.P_std + self.P_mean)

  def loss_weight(self, sigma):
    return (jnp.square(sigma) + jnp.square(self.sigma_data)) / jnp.square(
      sigma * self.sigma_data
    )

  def skip_scaling(self, sigma):
    return self.sigma_data**2 / (sigma**2 + self.sigma_data**2)

  def out_scaling(self, sigma):
    return sigma * self.sigma_data / (sigma**2 + self.sigma_data**2) ** 0.5

  def in_scaling(self, sigma):
    return 1 / (sigma**2 + self.sigma_data**2) ** 0.5

  def noise_conditioning(self, sigma):
    return 0.25 * jnp.log(sigma)

  def sampling_sigmas(self, num_steps):
    rho_inv = 1 / self.rho
    step_idxs = jnp.arange(num_steps, dtype=jnp.float32)
    sigmas = (
      self.sigma_max**rho_inv
      + step_idxs
      / (num_steps - 1)
      * (self.sigma_min**rho_inv - self.sigma_max**rho_inv)
    ) ** self.rho
    return jnp.concatenate([sigmas, jnp.zeros_like(sigmas[:1])])

  def sigma_hat(self, sigma, num_steps):
    gamma = (
      jnp.minimum(self.S_churn / num_steps, 2**0.5 - 1)
      if self.S_min <= sigma <= self.S_max
      else 0
    )
    return sigma + gamma * sigma


@dataclasses.dataclass
class EDMConfig:
  n_sampling_steps: int = 25
  sampler: str = "heun"
  parameterization: EDMParameterization = dataclasses.field(
    default_factory=EDMParameterization
  )


def edm(config: EDMConfig):
  """Construct denoising score-matching functions.

  Uses the EDM parameterization.

  Args:
    config: a EDMConfig object

  Returns:
    returns a tuple consisting of train_step, val_step and sampling functions
  """
  parameterization = config.parameterization

  def denoise(model, rng_key, inputs, sigma, context):
    new_shape = (-1,) + tuple(np.ones(inputs.ndim - 1, dtype=np.int32).tolist())
    inputs_t = inputs * parameterization.in_scaling(sigma).reshape(new_shape)
    noise_cond = parameterization.noise_conditioning(sigma)
    outputs = model(
      inputs=inputs_t,
      context=context,
      times=noise_cond,
    )
    skip = inputs * parameterization.skip_scaling(sigma).reshape(new_shape)
    outputs = outputs * parameterization.out_scaling(sigma).reshape(new_shape)
    outputs = skip + outputs
    return outputs

  def loss_fn(model, rng_key, batch):
    inputs = batch["inputs"]
    new_shape = (-1,) + tuple(np.ones(inputs.ndim - 1, dtype=np.int32).tolist())

    epsilon_key, noise_key, rng_key = jr.split(rng_key, 3)
    epsilon = jr.normal(epsilon_key, (inputs.shape[0],))
    sigma = parameterization.sigma(epsilon)

    noise = jr.normal(noise_key, inputs.shape) * sigma.reshape(new_shape)
    denoise_key, rng_key = jr.split(rng_key)
    target_hat = denoise(
      model,
      denoise_key,
      inputs=inputs + noise,
      sigma=sigma,
      context=batch.get("context"),
    )

    loss = jnp.square(inputs - target_hat)
    loss = parameterization.loss_weight(sigma).reshape(new_shape) * loss
    return loss.mean()

  def train_step(model, rng_key, batch, **kwargs):
    return nnx.value_and_grad(loss_fn)(model, rng_key, batch)

  def val_step(model, rng_key, batch, **kwargs):
    return loss_fn(model, rng_key, batch)

  sampler = getattr(samplers, config.sampler + "sample_fn")(config)
  return train_step, val_step, sampler



from collections.abc import Callable, Iterable

import jax
import wandb
from absl import logging
from flax import nnx
from jax import random as jr


# ruff: noqa: ANN001, ANN202, ANN003
def _step_and_val_fns(fns):
  step, eval = fns

  def _train_step(model, rng_key, optimizer, metrics, batch, **kwargs):
    model.train()
    loss, grads = step(model, rng_key, batch, **kwargs)
    optimizer.update(grads)
    metrics.update(loss=loss)
    return {"loss": loss}

  def _eval_step(model, rng_key, metrics, batch, **kwargs):
    model.eval()
    loss = eval(model, rng_key, batch, **kwargs)
    metrics.update(loss=loss)
    return {"loss": loss}

  return _train_step, _eval_step


# ruff: noqa: PLW2901,PLR0913
def train_fn(
  *,
  fns: tuple[Callable, Callable],
  shardings: tuple[jax.NamedSharding, jax.NamedSharding] | None = None,
  n_steps: int,
  eval_every_n_steps: int,
  n_eval_batches: int,
  log_to_wandb: bool = False,
  hooks: Iterable[Callable] = (),
) -> Callable:
  """Construct a function to train NNX models.

  Args:
    fns: a tuple of two callables. The first one is used as a step function
      , i.e., function to do gradient steps. The second one is used as an
      validation function.
    shardings: a tuple of shardings, the first one for the model, the second
      one for the data.
    n_steps: number of training/gradient steps
    eval_every_n_steps: specified how often to compute validation statistics.
    n_eval_batches: number of batches to use for validation
    log_to_wandb: whether to log results to wandb or not
    hooks: iterable of hooks

  Returns:
    returns a callable for training
  """

  def train(
    rng_key: jax.Array,
    model: nnx.Module,
    optimizer: nnx.Optimizer,
    train_itr: Iterable,
    val_itr: Iterable,
  ) -> None:
    """Train a NNX model.

    Args:
      rng_key: a jax.random.key object
      model: a NNX model
      optimizer: a nnx.Optimizer object
      train_itr: an infinite data loader, i.e., an iteratlor that keeps running.
        You can, for instance, construct this as a tfds.NumpyIterator or a
        grain.DataLoader.
      val_itr: an infinite data loader, i.e., an iteratlor that keeps running.
        You can, for instance, construct this as a tfds.NumpyIterator or a
        grain.DataLoader.
    """
    # get train and val fns
    step_fn, eval_fn = _step_and_val_fns(fns)
    # get model and replicate
    state = nnx.state((model, optimizer))
    if shardings is not None:
      state = jax.device_put(state, shardings[0])
      nnx.update((model, optimizer), state)
    # metrics
    metrics = nnx.MultiMetric(loss=nnx.metrics.Average("loss"))
    metrics_history = {}
    # run training
    step_key, rng_key = jr.split(rng_key)
    for step, batch in zip(range(1, n_steps + 1), train_itr):
      train_key, val_key = jr.split(jr.fold_in(step_key, step))
      if shardings is not None:
        batch = jax.device_put(batch, shardings[1])
      # do a gradient step
      step_fn(
        model=model,
        rng_key=train_key,
        optimizer=optimizer,
        metrics=metrics,
        batch=batch,
      )
      is_first_step = step == 1
      is_last_step = step == n_steps
      is_first_or_last_step = is_first_step or is_last_step
      if step % eval_every_n_steps == 0 or is_first_or_last_step:
        # store training losses
        for metric, value in metrics.compute().items():
          metrics_history[f"train/{metric}"] = float(value)
        # do evaluation loop
        for val_idx, batch in zip(range(n_eval_batches), val_itr):
          if shardings is not None:
            batch = jax.device_put(batch, shardings[1])
          eval_fn(
            model=model,
            rng_key=jr.fold_in(val_key, val_idx),
            metrics=metrics,
            batch=batch,
          )
        # store val losses
        for metric, value in metrics.compute().items():
          metrics_history[f"val/{metric}"] = float(value)
        metrics.reset()
        # log losses after each val round
        if jax.process_index() == 0:
          logging.info(
            f"loss at step {step}: "
            f"{metrics_history['train/loss']}/"
            f"{metrics_history['val/loss']}"
          )
        if log_to_wandb and jax.process_index() == 0:
          wandb.log(metrics_history, step=step)
      for h in hooks:
        h(step, model=model, optimizer=optimizer, metrics=metrics_history)

  return train



import dataclasses

import numpy as np
from flax import nnx
from jax import numpy as jnp
from jax import random as jr

from blaxbird._src.experimental import samplers


def _forward_process(inputs, times, noise):
  new_shape = (-1,) + tuple(np.ones(inputs.ndim - 1, dtype=np.int32).tolist())
  times = times.reshape(new_shape)
  inputs_t = times * inputs + (1.0 - times) * noise
  return inputs_t


@dataclasses.dataclass
class RFMParameterization:
  t_eps: float = 1e-5
  t_max: float = 1.0

  def sigma(self, eps):
    return self.t_eps + (self.t_max - self.t_eps)

  def loss_weight(self, t):
    return 1.0

  def skip_scaling(self, t):
    return 0.0

  def out_scaling(self, t):
    return 1.0

  def in_scaling(self, t):
    return 1.0

  def noise_conditioning(self, t):
    return t

  def sampling_sigmas(self, num_steps):
    return jnp.linspace(self.t_eps, self.t_max, num_steps)

  def sigma_hat(self, t, num_steps):
    return t


@dataclasses.dataclass
class RFMConfig:
  n_sampling_steps: int = 25
  sampler: str = "euler"
  parameterization: RFMParameterization = dataclasses.field(
    default_factory=RFMParameterization
  )


def rfm(config: RFMConfig = RFMConfig()):
  """Construct rectified flow matching functions.

  Args:
    config: a FlowMatchingConfig object

  Returns:
    returns a tuple consisting of train_step, val_step and sampling functions
  """
  parameterization = config.parameterization

  def _loss_fn(model, rng_key, batch):
    inputs = batch["inputs"]
    time_key, rng_key = jr.split(rng_key)
    times = jr.uniform(time_key, shape=(inputs.shape[0],))
    times = (
      times * (parameterization.t_max - parameterization.t_eps)
      + parameterization.t_eps
    )
    noise_key, rng_key = jr.split(rng_key)
    noise = jr.normal(noise_key, inputs.shape)
    inputs_t = _forward_process(inputs, times, noise)
    vt = model(inputs=inputs_t, times=times, context=batch.get("context"))
    ut = inputs - noise
    loss = jnp.mean(jnp.square(ut - vt))
    return loss

  def train_step(model, rng_key, batch, **kwargs):
    return nnx.value_and_grad(_loss_fn)(model, rng_key, batch)

  def val_step(model, rng_key, batch, **kwargs):
    return _loss_fn(model, rng_key, batch)

  sampler = getattr(samplers, config.sampler + "_sample_fn")(config)
  return train_step, val_step, sampler



import tensorflow as tf
import tensorflow_datasets as tfds
from jax import numpy as jnp
from jax import random as jr


def data_loaders(
  rng_key,
  outfolder,
  *,
  batch_size=128,
  buffer_size=1,
  prefetch_size=1,
  shuffle=True,
  split="train",
):
  datasets = tfds.load(
    "cifar10",
    try_gcs=False,
    split=split,
    data_dir=outfolder,
  )
  if isinstance(split, str):
    datasets = [datasets]
  itrs = []
  if isinstance(shuffle, bool):
    shuffle = [shuffle]
  assert len(datasets) == len(shuffle)
  for dataset, shuffle_me in zip(datasets, shuffle):
    itr_key, rng_key = jr.split(rng_key)
    itr = as_iterable(
      itr_key, dataset, batch_size, buffer_size, prefetch_size, shuffle_me
    )
    itrs.append(itr)
  return itrs


def as_iterable(rng_key, itr, batch_size, buffer_size, prefetch_size, shuffle):
  def process_fn(batch):
    img = tf.cast(batch["image"], tf.float32) / 255.0
    img = 2.0 * img - 1.0
    return {"inputs": img, "context": batch["label"]}

  max_int32 = jnp.iinfo(jnp.int32).max
  seed = jr.randint(rng_key, shape=(), minval=0, maxval=max_int32)
  return (
    itr.repeat()
    .shuffle(
      buffer_size,
      reshuffle_each_iteration=shuffle,
      seed=int(seed),
    )
    .map(process_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)
    .batch(batch_size, drop_remainder=True)
    .prefetch(prefetch_size)
    .as_numpy_iterator()
  )
